/*
 * Mate's Safari Extension v1.1.0
 * Created: 03/19
 * Last updated: 28/03/19
 * Twopeople Software (c)
 *
 * Required CSS files:
 *  "res/styles/pages/common/main.css"
 *  "res/styles/ui_components/tooltip/simple.css"
 *  "res/styles/ui_components/tooltip/help.css"
 *  "res/styles/ui_components/tooltip/helpSelected.css"
 *
 * Required JS files:
 *  "src/lib/jquery.js"
 *  "src/ui_components/scrollbar/scrollbar.js"
 */

//
// Constants
//

const TOOLTIP_PREFIX = "TnITTtw-";
const MAX_TEXT_LEN = 20000;
const CHROME_DEBUG = false; // set to false in production
const SAFARI_DEBUG = false;

// After how much time (in ms) show the floating button if not clicked
const FLOATING_BUTTON_HIDE_TIMEOUT = 2000;

const SYMBOL_LANGS = ['zh-TW', 'zh-CN'];

const GOOD_GOOGLE_LANGS = ["ar", "bn", "zh-CN", "zh-TW", "cs", "da", "nl", "en", "en-us", "et", "tl", "fi", "fr", "de", "hi", "id",
                           "it", "ja", "ko", "la", "pl", "pt", "ru", "sr", "si", "sk", "es", "su", "sv", "th", "uk", "vi", "cy"];

const GOOD_BING_LANGS = ["en", "bg", "hu", "vi", "el", "da", "he", "iw", "id", "es", "it", "yue", "ca", "zh-CN", "zh-TW", "ko", "ms", "de",
                         "da", "no", "pl", "ro", "ru", "sk", "sl", "th", "ta", "fi", "fr", "hr", "sv", "ja"];

const BAD_GOOGLE_LANGS = ["af", "sq", "hy", "bs", "ca", "hr", "eo", "el", "hu", "is", "jw", "km", "lv", "mk", "ml",
                          "mr", "my", "ne", "no", "ro", "sw", "ta", "te", "tr"];

const BAD_BING_LANGS = ["ar", "pt", "hi", "cs"];

const LOCALES = {
en: {
    'Tooltip_NetflixSave': 'SAVE',
    'Tooltip_NetflixContinue': 'CONTINUE',
    "Kernel_IPA": "Phonetic transcription",
    "CommonUi_LangPart_Transliteration": "Transliteration",
    "Window_Offline": "No Internet Connection",
    "CommonUi_LangPart_Noun": "Nouns",
    "CommonUi_LangPart_Verb": "Verbs",
    "CommonUi_LangPart_Adjective": "Adjectives",
    "CommonUi_LangPart_Adverb": "Adverbs",
    "CommonUi_LangPart_Pronoun": "Pronouns",
    "CommonUi_LangPart_Preposition": "Prepositions",
    "CommonUi_LangPart_Conjunction": "Conjunctions",
    "CommonUi_LangPart_Interjection": "Interjections",
    "CommonUi_LangPart_Abbreviation": "Abbreviations",
    "CommonUi_LangPart_Phrase": "Phrases",
    "CommonUi_LangPart_Suffix": "Suffixes",
    "CommonUi_LangPart_Auxiliaryverb": "Auxiliary Verbs",
    'FullPageTranslatedTo': 'Translated to',
    "FullPageShowOriginal": "Show original",
    "FullPageTranslate": "Translate"
}
};

const CURRENT_LOCALE = window.navigator.language
.split('-').shift()
.split('_').shift();

//
// Flags defined at start
//

var dark_mode = false;
var double_click_enabled = true;
var selection_enabled = true;
var show_translit = true;
var show_ipa = true;
var scale = 1.0; // possible options: 1.0 - big (default), 0.85 - medium, 0.75 - small

//
// Dynamically changed flags
//

var tooltip_id = 0;
var is_translating = false;
var skip_selection_translation = false;
var last_call_args = {
selectionBB: null,
from: "",
to: "",
text: ""
};

//
// On DOM loaded
//

$(function () {
  $(window).dblclick(initTranslationOnDoubleClick);
  $(window).mouseup(function (event) {
                    // to prevent firing alongside with the double click listener
                    setTimeout(function () {
                               initTranslationOnSelection(event);
                               }, 10);
                    });
  });

function getLocale(key) {
    return (LOCALES[CURRENT_LOCALE] || LOCALES["en"])[key] || "";
}

if (CHROME_DEBUG) {
    function processCall() {
        var p = [];
        $.each(arguments, function (k, v) {
               p.push(v);
               });
        return p.join(',');
    }
    
    var safari = {
    callback: function () {
    },
        
    self: {
    addEventListener: function (msg, callback) {
        safari.callback = callback;
    }
    },
        
    extension: {
    baseURI: 'res/images/flags/',
        
    dispatchMessage: function (msg, params) {
        chrome.runtime.sendMessage({
                                   action: processCall('app', 'option', params.action),
                                   params: params
                                   }, function (data) {
                                   safari.callback(data.old_data.params);
                                   });
    }
    }
    };
}

function hasFocusedInputs() {
    var has_focus_elements = false;
    
    $(document.body).find("*:focus").each(function () {
                                          if ($(this).is("input,textarea") || this.contentEditable == true) {
                                          has_focus_elements = true;
                                          }
                                          });
    
    return has_focus_elements;
}

function _getSelection() {
    var selection = null;
    
    if (window.getSelection) {
        selection = window.getSelection();
    } else if (document.getSelection) {
        selection = document.getSelection();
    }
    
    return selection;
}

function getSelectedText() {
    var selection = _getSelection();
    
    return selection
    ? selection.toString().trim()
    : '';
}

function initTranslationOnDoubleClick(event) {
    if (!double_click_enabled || (event.target.className || '').indexOf(TOOLTIP_PREFIX) > -1
        || hasFocusedInputs() || !getSelectedText()) {
        return;
    }
    
    skip_selection_translation = true;
    
    console.log('double click');
    
    handleTranslation("double-click");
}

// @returns null
function initTranslationOnSelection(event) {
    if (!selection_enabled || hasFocusedInputs() || !getSelectedText() || (event.target.className || '').indexOf(TOOLTIP_PREFIX) > -1) {
        //
        // hide button and abort
        //
        hideSelectionButton();
        return;
    }
    
    if (skip_selection_translation) {
        skip_selection_translation = false;
        return;
    }
    
    console.log('select');
    
    var selection = _getSelection();
    var params = getSelectionParameters(selection);
    
    last_call_args.text = selection.toString();
    last_call_args.selectionBB = params;
    
    $('.' + TOOLTIP_PREFIX + 'translate-selection-button').remove();
    
    var height = 24;
    var width = height;
    var x = params.left + params.width / 2 + window.pageXOffset - width / 2;
    var y;
    
    if (params.top + params.height + 10 + height > window.innerHeight) {
        y = params.top - 10 - height + window.pageYOffset; // floating button is above
    } else {
        y = params.top + params.height + 10 + window.pageYOffset; // floating button is below
    }
    
    $('body').append(
                     $('<div class="' + TOOLTIP_PREFIX + 'translate-selection-button"></div>')
                     .css({
                          top: y,
                          left: x
                          })
                     .on('click', function (event) {
                         event.stopPropagation();
                         event.preventDefault();
                         
                         handleTranslation('selection');
                         })
                     );
    
    setTimeout(function () {
               $('.' + TOOLTIP_PREFIX + 'translate-selection-button').fadeOut(250, function () {
                                                                              $(this).remove();
                                                                              });
               }, FLOATING_BUTTON_HIDE_TIMEOUT);
}

// @returns null
function hideSelectionButton() {
    $('.' + TOOLTIP_PREFIX + 'translate-selection-button').fadeOut(150, function () {
                                                                   $(this).remove();
                                                                   });
}

//
// Handling responses from the parent app
//

safari.self.addEventListener("message", function (data) {
                             if (!CHROME_DEBUG) {
                             // The data here is the whole event object,
                             // data.message is what's actually being sent from the native code
                             data = data.message;
                             }
                             
                             if (SAFARI_DEBUG) {
                             console.log('Message:', data);
                             }
                             
                             if (data.action === "translate") {
                             displayTranslation(data);
                             }
                             
                             if ((data.action === 'run_tts' || data.action === 'stop_tts') && data.finished) {
                             onAudioPlaybackFinished(data);
                             }
                             
                             if (data.action === "get_show_translit_option") {
                             show_translit = data.value;
                             }
                             
                             if (data.action === "get_show_ipa_option") {
                             show_ipa = data.value;
                             }
                             
                             if (data.action === "get_dark_mode_option") {
                             dark_mode = data.value;
                             }
                             
                             if (data.action === "get_double_click_option") {
                             double_click_enabled = data.value;
                             }
                             
                             if (data.action === "get_selection_option") {
                             selection_enabled = data.value;
                             }
                             
                             if (data.action === "get_scale_option") {
                             scale = +data.value;
                             }
                             
                             if (data.action === "get_selected_text") {
                             safari.extension.dispatchMessage("message", {
                                                              action: "get_selected_text",
                                                              text: getSelectedText(),
                                                              url: document.location.href
                                                              });
                             }
                             
                             if (data.action === "translate_selection") {
                                handleTranslation('selection');
                             }
                             
});



//
// Request all settings on start
// to show the first translation properly, too
//

requestSettingsFromParentApp();

if (document.location.href.indexOf('https://www.netflix.com/watch') > -1) {
    safari.extension.dispatchMessage("message", {
                                     "action": "netflix"
                                     });
}

function requestSettingsFromParentApp() {
    safari.extension.dispatchMessage("message", {
                                     "action": "get_dark_mode_option"
                                     });
    
    safari.extension.dispatchMessage("message", {
                                     "action": "get_show_translit_option"
                                     });
    
    safari.extension.dispatchMessage("message", {
                                     "action": "get_show_ipa_option"
                                     });
    
    safari.extension.dispatchMessage("message", {
                                     "action": "get_double_click_option"
                                     });
    
    safari.extension.dispatchMessage("message", {
                                     "action": "get_selection_option"
                                     });
    
    // ready to use
    // as soon as there's an option in settings for this
    /*safari.extension.dispatchMessage("message", {
     "action": "get_scale_option"
     });*/
}

// @returns null
function handleTranslation(type) {
    //
    // Fetch settings on each event to stay up to date
    // I.e. not only on page reload (as it was before)
    //
    
    requestSettingsFromParentApp();
    
    //
    // Carry on with translation:
    //
    
    if (is_translating) {
        return;
    }
    
    var id;
    var selection = _getSelection();
    var selected_text = selection.toString();
    
    if (selected_text) {
        last_call_args.selectionBB = selection.getRangeAt(0).getBoundingClientRect();
        last_call_args.text = selected_text;
        //last_call_args.from = from;
        //last_call_args.to = to;
        
        id = showTooltip(selection);
        
        toggleLoadingInTooltip(id, true);
    } else if (last_call_args.selectionBB && last_call_args.text) {
        //last_call_args.from = from;
        //last_call_args.to = to;
        
        id = showTooltip(last_call_args.selectionBB);
        toggleLoadingInTooltip(id, true);
        
        selected_text = last_call_args.text;
    } else {
        return;
    }
    
    is_translating = true;
    
    safari.extension.dispatchMessage("message", {
                                     "action": "translate",
                                     "text": selected_text.substr(0, MAX_TEXT_LEN),
                                     "id": id,
                                     "type": type,
                                     "source": document.location.href
                                     });
}

function setupScroll(he, h, vis, ent, db, t, trs, ot, speed) {
    var willShow = parseInt($(he).css('height')) >= h;
    
    $('#' + trs)[$(he).get(0) ? (willShow ? 'show' : 'hide') : 'hide']();
    
    var scrollInstance = new dw_scrollObj(vis, ent);
    scrollInstance.setUpScrollbar(db, t, 'v', 1, 1);
    scrollInstance.setUpScrollControls(trs);
    
    if (ot) {
        dw_scrollObj.scrollTo(vis, 1, ot, speed || 250);
    }
    
    return willShow;
}

// @returns null
function displayTranslation(data) {
    $('.' + TOOLTIP_PREFIX + 'tooltip-' + data.id)
    .find('.' + TOOLTIP_PREFIX + 'listen-original')
    .attr('data-from', data.from);
    
    is_translating = false;
    
    if (data.offline) {
        toggleOfflineInTooltip(data.id, true);
        return;
    }
    
    data.translation = JSON.parse(data.translation);
    
    toggleLoadingInTooltip(data.id, false);
    setTooltipContents(data.id, wrapper.wrap(data.translation));
    
    $('.' + TOOLTIP_PREFIX + 'tooltip-' + data.id)
    .find('.' + TOOLTIP_PREFIX + 'listen-translation')
    .attr('data-to', data.to);
    
    var $tt = $('.' + TOOLTIP_PREFIX + 'tooltip-' + data.id);
    
    if ($tt.find('.' + TOOLTIP_PREFIX + 'top-arr0w').is(':visible')) {
        $tt.find('#' + TOOLTIP_PREFIX + 'tr-scrollbar').addClass(TOOLTIP_PREFIX + 'top-scroll');
    }
    
    $tt.find('.' + TOOLTIP_PREFIX + 'from-flag').attr('src', safari.extension.baseURI + data.from + '@2x.png');
    $tt.find('.' + TOOLTIP_PREFIX + 'to-flag').attr('src', safari.extension.baseURI + data.to + '@2x.png');
    
    if (!data.translation[0] && data.translation[3].length < 35) {
        $tt.find('.' + TOOLTIP_PREFIX + 'padded-single-translation').addClass(TOOLTIP_PREFIX + 'short-padded-single-translation');
    }
    
    if (SYMBOL_LANGS.indexOf(data.to) > -1) {
        $tt.find('.' + TOOLTIP_PREFIX + 'content-layout').addClass(TOOLTIP_PREFIX + 'non-bold-contents');
    } else {
        $tt.find('.' + TOOLTIP_PREFIX + 'content-layout').removeClass(TOOLTIP_PREFIX + 'non-bold-contents');
    }
    
    ctrlTooltipOrigVisibility(false, data.from, $tt);
    ctrlTooltipTransVisibility(false, data.to, $tt);
    ctrlSynonymVis(false, data.to, $tt);
    
    $tt.find('.' + TOOLTIP_PREFIX + 'listen-original').on('click', playTooltip);
    $tt.find('.' + TOOLTIP_PREFIX + 'listen-butt0n').on('click', playTooltip);
    $tt.find('.' + TOOLTIP_PREFIX + 'listen-v-item').on('click', playTooltipSynonym);
    $tt.find('.' + TOOLTIP_PREFIX + 'show-reversed')
    .find('.' + TOOLTIP_PREFIX + 'flag')
    .on('click', translateReverse);
    $tt.find('.' + TOOLTIP_PREFIX + 'unpin').on('click', unpin);
    $tt.find('.' + TOOLTIP_PREFIX + 'small-copy-button').on('click', copySynonym);
    $tt.find('.' + TOOLTIP_PREFIX + 'copy-translation-butt0n')
    .on('click', copyMainTranslation)
    .attr('data-tid', data.id);
    
    fadeInTooltip(data.id, function () {
                  setupScroll('.' + TOOLTIP_PREFIX + 'content-layout-' + data.id, 238,
                              TOOLTIP_PREFIX + 'trVisibleLayout-' + data.id,
                              TOOLTIP_PREFIX + 'trEntireLayout-' + data.id,
                              TOOLTIP_PREFIX + 'dragBar-' + data.id,
                              TOOLTIP_PREFIX + 'track-' + data.id,
                              TOOLTIP_PREFIX + 'tr-scrollbar-' + data.id);
                  });
    
    setTimeout(hideSelectionButton, 250);
}

function fadeInTooltip(ttid, callback) {
    $('.' + TOOLTIP_PREFIX + 'tooltip-' + ttid).fadeIn(400, callback);
}

function copyMainTranslation() {
    var $this = $(this);
    
    if ($this.hasClass(TOOLTIP_PREFIX + 'copied')) {
        return;
    }
    
    $this.addClass(TOOLTIP_PREFIX + 'copied');
    setTimeout(function () {
               $this.removeClass(TOOLTIP_PREFIX + 'copied');
               }, 2500);
    
    var text = getListenValue('trans', $('.' + TOOLTIP_PREFIX + 'tooltip-' + $(event.target).data('tid')));
    
    copyToClipboard(text);
}

function translateReverse() {
    var action = "";
    
    if ($(this).hasClass(TOOLTIP_PREFIX + 'from-flag')) {
        action = "open_from_lang_picker";
    } else {
        action = "open_to_lang_picker";
    }
    
    safari.extension.dispatchMessage("message", {
                                     action: action
                                     });
}

function playTooltipSynonym(event) {
    var val = $(this).parent().find('.' + TOOLTIP_PREFIX + 'main-of-item').html();
    
    playUniversal(
                  $(this),
                  'lang:' + $(this).data('langto'),
                  val,
                  ['ctrlSynonymVis', 'ctrlTooltipOrigVisibility',
                   'ctrlTooltipTransVisibility'],
                  null,
                  event
                  );
}

function playTooltip(e) {
    e.stopPropagation();
    
    if ($(this).hasClass(TOOLTIP_PREFIX + 'listen-original')) {
        playTooltipOriginal.call(this, e, $(this).data('from'));
    } else {
        playTooltipTranslation.call(this, e, $(this).data('to'));
    }
}

function playTooltipOriginal(event, lang) {
    playUniversal(
                  $(this),
                  'lang:' + lang,
                  getListenValue('orig', event),
                  ['ctrlTooltipTransVisibility', 'ctrlSynonymVis'],
                  null,
                  event
                  );
}

function playTooltipTranslation(event, lang) {
    playUniversal(
                  $(this),
                  'lang:' + lang,
                  getListenValue('trans', event),
                  ['ctrlTooltipOrigVisibility', 'ctrlSynonymVis'],
                  null,
                  event
                  );
}

var listen_target_id = 0;
var listen_targets = {};

function playUniversal($button, dir, input, vis_fns, dl, ctx) {
    if (!input || $button.is('[class$="listen-disabled"]')) {
        return;
    }
    
    var lang = dl || (dir.substr(0, 5) === 'lang:' ? dir.substr(5) : 'en');
    
    if ($button.hasClass('stop-audio')) {
        safari.extension.dispatchMessage("message", {
                                         action: "stop_tts",
                                         lang: lang,
                                         vis_fns: vis_fns,
                                         target_id: +$button.data('tid')
                                         });
    } else {
        $button.addClass('stop-audio');
        
        vis_fns.forEach(function (vis_fn) {
                        window[vis_fn](true, lang, ctx);
                        });
        
        var target_id = ++listen_target_id;
        $button.attr('data-tid', target_id);
        listen_targets[target_id] = ctx;
        
        safari.extension.dispatchMessage("message", {
                                         action: "run_tts",
                                         vis_fns: vis_fns,
                                         text: input.trim(),
                                         lang: lang,
                                         target_id: target_id
                                         });
    }
}

function onAudioPlaybackFinished(data) {
    if (!data) {
        return;
    }
    
    $(listen_targets[data.target_id].target).removeClass('stop-audio');
    
    data.vis_fns.forEach(function (vis_fn) {
                         window[vis_fn](
                                        false,
                                        data.lang,
                                        listen_targets[data.target_id]
                                        );
                         });
    
    delete listen_targets[data.target_id];
}

function unpin(event) {
    var $tooltip = $(getTooltipWrapRecursively(event.target));
    
    $tooltip.addClass(TOOLTIP_PREFIX + 'unpinned');
    $tooltip.find('.' + TOOLTIP_PREFIX + 'arr0w').remove();
    
    //
    //
    // Update the contents of tooltip
    
    var time = 175;
    $tooltip.find('.' + TOOLTIP_PREFIX + 'utils').slideUp(time);
    setTimeout(function () {
               $tooltip.find('.' + TOOLTIP_PREFIX + 'unpinned-utils').css('display', 'block');
               $tooltip.find('.' + TOOLTIP_PREFIX + 'original-wrap').slideDown(time, function () {
                                                                               var ttid = $tooltip.data('ttid');
                                                                               
                                                                               $(this).css('display', 'block');
                                                                               
                                                                               setupScroll('.' + TOOLTIP_PREFIX + 'content-layout-' + ttid, 238,
                                                                                           TOOLTIP_PREFIX + 'trVisibleLayout-' + ttid,
                                                                                           TOOLTIP_PREFIX + 'trEntireLayout-' + ttid,
                                                                                           TOOLTIP_PREFIX + 'dragBar-' + ttid,
                                                                                           TOOLTIP_PREFIX + 'track-' + ttid,
                                                                                           TOOLTIP_PREFIX + 'tr-scrollbar-' + ttid);
                                                                               
                                                                               // TODO
                                                                               // resize tooltip (without twitches desirably)
                                                                               // resizeTooltip(ttid, true);
                                                                               });
               }, time - 75);
    
    //
    //
    // Dragging
    
    var startPosXInTt = 0, startPosYInTt = 0;
    
    var tt_move = function (event) {
        $tooltip.css({
                     left: event.clientX - startPosXInTt,
                     top: event.clientY - startPosYInTt
                     });
    };
    
    $tooltip.find('.' + TOOLTIP_PREFIX + 'unpinned-utils').on('mousedown', function (event) {
                                                              if (event.which === 1) {
                                                              startPosXInTt = event.clientX - parseInt($tooltip.css('left'));
                                                              startPosYInTt = event.clientY - parseInt($tooltip.css('top'));
                                                              
                                                              $(window).on('mousemove', tt_move);
                                                              }
                                                              });
    $(window).on('mouseup', function () {
                 $(window).off('mousemove', tt_move);
                 });
    
    //
    //
    // Closing
    
    $('.' + TOOLTIP_PREFIX + 'close-unpinned').on('click', function (event) {
                                                  var $to_close_tooltip = $(getTooltipWrapRecursively(event.target));
                                                  $to_close_tooltip.fadeOut(125, function () {
                                                                            $(this).remove();
                                                                            });
                                                  });
}

function copySynonym() {
    var prefix = TOOLTIP_PREFIX;
    
    var $this = $(this);
    var _class = prefix + 'copied';
    
    if ($this.hasClass(_class)) {
        return;
    }
    
    var $el = $(this).parent().find('.' + prefix + 'main-of-item');
    
    $this.addClass(_class);
    setTimeout(function () {
               $this.removeClass(_class);
               }, 2500);
    
    copyToClipboard($el.html());
}

function copyToClipboard(text) {
    var $cc = $('<textarea>')
    .css({
         position: 'absolute',
         top: -1000,
         left: -1000
         })
    .appendTo('body');
    $cc.val(text).focus().select();
    document.execCommand('Copy');
    $cc.remove();
}

function getTooltipWrapRecursively(target) {
    return $(target).attr('id') == TOOLTIP_PREFIX + 'tooltip-wrap'
    ? target
    : getTooltipWrapRecursively($(target).parent().get(0));
}

function isUtterable(lang) {
    return GOOD_GOOGLE_LANGS.indexOf(lang) > -1
    || GOOD_BING_LANGS.indexOf(lang) > -1
    || BAD_GOOGLE_LANGS.indexOf(lang) > -1
    || BAD_BING_LANGS.indexOf(lang) > -1;
}

function ctrlTooltipOrigVisibility(playing, lang, e) {
    ctrlTooltipPlaybackButtonVisibility(playing, lang, e, 'original', 'from', 'orig');
}

function ctrlTooltipTransVisibility(playing, lang, e) {
    ctrlTooltipPlaybackButtonVisibility(playing, lang, e, 'translation', 'to', 'trans');
}

function ctrlTooltipPlaybackButtonVisibility(playing, lang, e, t_btn, t_dir, t_val) {
    var $context;
    
    if (e.target) {
        $context = $(getTooltipWrapRecursively(e.target));
    } else if (e.innerHTML) {
        $context = $(e);
    } else {
        $context = e;
    }
    
    var to_lang = $context.find('.' + TOOLTIP_PREFIX + 'listen-' + t_btn).data(t_dir);
    
    if (!isUtterable(to_lang)) {
        $context.addClass(TOOLTIP_PREFIX + 'no-' + t_val + '-tts');
    } else {
        $context.removeClass(TOOLTIP_PREFIX + 'no-' + t_val + '-tts');
    }
    
    var allowed = !playing && getListenValue(t_val, e);
    
    $context.find('.' + TOOLTIP_PREFIX + 'listen-' + t_btn)[(allowed ? 'remove' : 'add') + 'Class'](TOOLTIP_PREFIX + 'listen-disabled');
}

function ctrlSynonymVis(playing, lang, event) {
    if (!event) {
        event = $("body");
    }
    
    var $context = event.target ? $(event.target)
    .parent()
    .parent()
    .parent()
    .parent() : event;
    
    var allowed = !playing;
    var prefix = TOOLTIP_PREFIX;
    
    if (!isUtterable(lang)) {
        $context.addClass(prefix + 'no-tts');
    } else {
        $context.removeClass(prefix + 'no-tts');
    }
    
    $('.' + prefix + 'listen-v-item').each(function () {
                                           var $this = $(this);
                                           if (!$this.hasClass('stop-audio')) {
                                           $this[(allowed ? 'remove' : 'add') + 'Class'](prefix + 'listen-disabled');
                                           }
                                           });
}

function getListenValue(s, event) {
    var $where = !event.target ? event : $(getTooltipWrapRecursively(event.target));
    
    if (s === 'orig') {
        var i = $where.find('.' + TOOLTIP_PREFIX + 'original-wrap .' + TOOLTIP_PREFIX + 'mv-text-part').html();
        if (!i) {
            return $where.find('.' + TOOLTIP_PREFIX + 'original-wrap .' + TOOLTIP_PREFIX + 'tpart').html();
        } else {
            return i;
        }
    } else if (s === 'trans') {
        var i = $where.find('.' + TOOLTIP_PREFIX + 'main-variant .' + TOOLTIP_PREFIX + 'mv-text-part').html();
        if (!i) {
            return $where.find('.' + TOOLTIP_PREFIX + 'trans-wrap .' + TOOLTIP_PREFIX + 'tpart').html();
        } else {
            return i;
        }
    }
    
    return '';
}

function getSelectionParameters(s) {
    return s.getRangeAt ? $.extend({
                                   scrollX: window.pageXOffset,
                                   scrollY: window.pageYOffset
                                   }, s.getRangeAt(0).getBoundingClientRect()) : {
    left: s.x || s.left,
    top: s.y || s.top,
    width: s.width,
    height: s.height
    };
}

function parseHtml(h) {
    return h.replace(/&lt;/gi, '<').replace(/&gt;/gi, '>');
}

function compileString(str, data) {
    data = data || {};
    str = parseHtml(str);
    
    for (var key in data) {
        str = str
        .replace(new RegExp("<%=" + key + "%>", 'g'), data[key])
        .replace(new RegExp("<%= " + key + " %>", 'g'), data[key]);
    }
    
    return str;
}

const TOOLTIP_CONTENTS_HTML = '\
<div class="<%=prefix%>t <%=prefix%>help-selected-wrap <%=prefix%>hsw-<%=ttid%>">\
<div class="<%=prefix%>t <%=prefix%>help-inside-layout <%=prefix%>hil-<%=ttid%>">\
<div class="<%=prefix%>unpinned-utils">\
<div class="<%=prefix%>pro-img"></div>\
<div class="<%=prefix%>close-unpinned"></div>\
</div>\
<div class="<%=prefix%>utils">\
<div title="<%=title_listen_original%>" class="<%=prefix%>util-butt0n <%=prefix%>listen-original" data-from="<%=from%>"></div>\
<div title="<%=title_unpin%>" class="<%=prefix%>util-butt0n <%=prefix%>unpin"></div>\
<div title="<%=title_show_reversed%>" class="<%=prefix%>show-reversed">\
<img class="<%=prefix%>flag <%=prefix%>from-flag">\
<div class="<%=prefix%>dir-arrow"></div>\
<img class="<%=prefix%>flag <%=prefix%>to-flag">\
</div>\
</div>\
<div class="<%=prefix%>trVisibleLayout" id="<%=prefix%>trVisibleLayout-<%=ttid%>">\
<div class="<%=prefix%>trEntireLayout" id="<%=prefix%>trEntireLayout-<%=ttid%>">\
<div class="<%=prefix%>t <%=prefix%>content-layout <%=prefix%>content-layout-<%=ttid%>">\
<%=content%></div>\
</div>\
</div>\
\
<div class="<%=prefix%>tr-scrollbar" id="<%=prefix%>tr-scrollbar-<%=ttid%>">\
<div class="<%=prefix%>track" id="<%=prefix%>track-<%=ttid%>">\
<div class="<%=prefix%>dragBar" id="<%=prefix%>dragBar-<%=ttid%>"></div>\
</div>\
</div>\
</div>\
\
<div class="<%=prefix%>netflix-buttons <%=prefix%>nf-<%=ttid%>">\
<div class="<%=prefix%>netflix-button <%=prefix%>netflix-save t-<%=ttid%>"><%=netflix_save%></div>\
<div class="<%=prefix%>netflix-button <%=prefix%>netflix-continue"><%=netflix_continue%></div>\
</div>\
\
<div class="<%=prefix%>info-warn <%=prefix%>ddopt-<%=ttid%> <%=prefix%>hide" data-option="double_click">\
<span title="<%=dd_localized_desc%>"><%=dd_localized%></span>\
<div class="<%=prefix%>close-info-warn <%=prefix%>close-dd-option" data-ttid="<%=ttid%>"></div>\
</div>\
<div class="<%=prefix%>info-warn <%=prefix%>selopt-<%=ttid%> <%=prefix%>hide" data-option="selection">\
<span title="<%=sel_localized_desc%>"><%=sel_localized%></span>\
<div class="<%=prefix%>close-info-warn <%=prefix%>close-sel-option" data-ttid="<%=ttid%>"></div>\
</div>\
\
<div class="<%=prefix%>info-warn <%=prefix%>iw-<%=ttid%> <%=prefix%>hide" data-option="localized">\
<a target="_blank" title="<%=ti_localized_desc%>"><%=ti_localized%></a>\
<div class="<%=prefix%>close-info-warn <%=prefix%>close-loc-option" data-ttid="<%=ttid%>"></div>\
</div>\
<div class="<%=prefix%>loading"><div class="<%=prefix%>mate-loading"></div></div>\
<div class="<%=prefix%>offline"><span><%=l_offline%></span></div>\
</div>\
';

const TOOLTIP_HTML = '\
<div class="<%=prefix%>tooltip-main-wrap <%=prefix%>tooltip-<%=ttid%> <%=prefix%>t" id="<%=prefix%>tooltip-wrap" data-ttid="<%=ttid%>">\
<div class="<%=prefix%>t <%=prefix%>inside-layout">\
<div class="<%=prefix%>t <%=prefix%>content">' + TOOLTIP_CONTENTS_HTML + '</div>\
</div>\
</div>\
';

function getOParamsInstance(g) {
    g.width = g.width || 0;
    g.height = g.height || 0;
    g.pl = g.pl || 0;
    g.pr = g.pr || 0;
    g.pt = g.pt || 0;
    g.pb = g.pb || 0;
    return g;
}

function closeTooltip(e) {
    if ((typeof e === 'boolean' && e) || (typeof e !== 'boolean' && !$(e.target).hasClass(TOOLTIP_PREFIX + 't'))) {
        $('body').unbind();
        
        var tooltip = $('.' + TOOLTIP_PREFIX + 'tooltip-main-wrap');
        
        if (tooltip.length > 0) {
            tooltip.each(function () {
                         if (!$(this).hasClass(TOOLTIP_PREFIX + 'unpinned')) {
                         $(this).remove();
                         }
                         });
        }
    }
}

$.fn.measure = function (fn) {
    var el = $(this).clone(false);
    el.css({
           visibility: 'hidden',
           position: 'absolute'
           });
    el.appendTo('body');
    var result = fn.apply(el);
    el.remove();
    return result;
};

function getBodyScrollLeft() {
    return document.documentElement.scrollLeft || document.body.scrollLeft || 0;
}

function getBodyScrollTop() {
    return document.documentElement.scrollTop || document.body.scrollTop || 0;
}

const Y_OFFSET = 10;

function computeTooltipPosition(el, ix, iy, oparams, scale, callback) {
    var pos = [0, 0, 'bottom'];
    
    var tooltip_width = 0;
    var tooltip_height = 0;
    
    $(el.get()).measure(function () {
                        tooltip_width = this.width();
                        tooltip_height = this.height();
                        
                        var absolute_selection_left_scroll = ix + getBodyScrollLeft();
                        var absolute_selection_top_scroll = iy + getBodyScrollTop();
                        
                        var selection_absolute_width = oparams.width + oparams.pl + oparams.pr;
                        var selection_absolute_height = oparams.height + oparams.pt + oparams.pb;
                        
                        pos[0] = absolute_selection_left_scroll - tooltip_width / 2 + selection_absolute_width / 2;
                        pos[1] = absolute_selection_top_scroll - tooltip_height - Y_OFFSET * scale;
                        pos[2] = 'bottom'; // tooltip is above, arrow is on bottom
                        
                        // Horizontal alignment
                        if (pos[0] - getBodyScrollLeft() < 1) {
                        
                        // stick to the left side
                        
                        pos[0] = getBodyScrollLeft() + 1 - tooltip_width * (1 - scale) / 2;
                        } else if (pos[0] + tooltip_width > document.body.clientWidth) {
                        
                        // stick to the right side
                        
                        pos[0] = document.body.clientWidth - tooltip_width - 1 + tooltip_width * (1 - scale) / 2;
                        }
                        
                        // A vertical one
                        if (pos[1] - getBodyScrollTop() < 1) {
                        pos[1] = absolute_selection_top_scroll + selection_absolute_height + Y_OFFSET * scale;
                        
                        // tooltip below the selection
                        // if scale != 1
                        pos[1] -= tooltip_height * (1 - scale) / 2;
                        
                        pos[2] = 'top'; // tooltip is below, arrow is on top
                        } else {
                        pos[1] += tooltip_height * (1 - scale) / 2;
                        }
                        
                        callback(pos);
                        });
}

// @returns id - tooltip ID
function showTooltip(selection) {
    var ttid = ++tooltip_id;
    var sel_params = getSelectionParameters(selection);
    
    closeTooltip(true);
    
    var $tooltip = $(compileString(TOOLTIP_HTML, {
                                   content: '',
                                   prefix: TOOLTIP_PREFIX,
                                   ttid: ttid,
                                   l_original: getLocale('Kernel_Original'),
                                   l_open: getLocale('Kernel_OpenGt'),
                                   l_reversed: getLocale('Kernel_Reverse'),
                                   l_unpin: getLocale('Kernel_Unpin'),
                                   l_highlight: getLocale('Kernel_Highlight'),
                                   l_loading: getLocale('Kernel_Loading'),
                                   l_offline: getLocale('Window_Offline'),
                                   title_highlight_button: getLocale('Tooltip_Highlight'),
                                   title_open_link: getLocale('Tootip_OpenLink'),
                                   title_listen_original: getLocale('Tooltip_ListenOriginal'),
                                   title_show_reversed: getLocale('Tooltip_ShowReversed'),
                                   title_unpin: getLocale('Tooltip_Unpin'),
                                   dd_localized: getLocale('Tooltip_DisableDoubleClick'),
                                   dd_localized_desc: getLocale('Tooltip_DisableDoubleClickDesc'),
                                   sel_localized: getLocale('Tooltip_DisableSelection'),
                                   sel_localized_desc: getLocale('Tooltip_DisableSelectionDesc'),
                                   netflix_save: getLocale('Tooltip_NetflixSave'),
                                   netflix_continue: getLocale('Tooltip_NetflixContinue')
                                   }));
    
    var maxZ = Math.max.apply(null,
                              $.map($('body *'), function (e, n) {
                                    if ($(e).css('position') != 'static')
                                    return parseInt($(e).css('z-index')) || 1;
                                    }));
    
    $tooltip.css({
                 "z-index": maxZ + 1,
                 maxWidth: 450,
                 maxHeight: 325
                 });
    
    $(window).on('click', closeTooltip);
    
    var left = 0;
    var top = 0;
    var params;
    
    left += sel_params.left; //- (window.scrollX - sel_params.scrollX);
    top += sel_params.top; //- (window.scrollY - sel_params.scrollY);
    params = getOParamsInstance(sel_params);
    
    $tooltip.css('transform', 'scaleX(' + scale + ') scaleY(' + scale + ')');
    
    computeTooltipPosition($tooltip, left, top, params, scale, function (tp) {
                           var real_y = tp[1] - 5;
                           
                           $tooltip.css({
                                        left: tp[0],
                                        top: real_y
                                        }).animate({
                                                   top: real_y
                                                   }, 300);
                           
                           var $dest = $('body');
                           var _dark_mode = dark_mode;
                           
                           if (document.location.href.indexOf('https://www.netflix.com/watch') > -1) {
                           $dest = $('.nf-player-container');
                           _dark_mode = true;
                           } else if ($dest.length === 0) {
                           $dest = $('html');
                           }
                           
                           $tooltip.data('ttid', ttid);
                           
                           if (_dark_mode) {
                           $tooltip.addClass(TOOLTIP_PREFIX + 'dark-mode');
                           }
                           
                           $dest.append($tooltip.get());
                           
                           attachArrows(ttid);
                           applyArrowPosition($tooltip, tp, left, top, scale, params);
                           
                           $('#' + TOOLTIP_PREFIX + 'tooltip-wrap *').addClass('TnITTtw-t');
                           });
    
    return ttid;
}

function attachArrows(ttid) {
    $('.' + TOOLTIP_PREFIX + 'tooltip-' + ttid).prepend(
                                                        $('<div>')
                                                        .addClass(TOOLTIP_PREFIX + 't').addClass(TOOLTIP_PREFIX + 'arr0w').addClass(TOOLTIP_PREFIX + 'top-arr0w')
                                                        .get()
                                                        );
    
    $('.' + TOOLTIP_PREFIX + 'tooltip-' + ttid).append(
                                                       $('<div>')
                                                       .addClass(TOOLTIP_PREFIX + 't').addClass(TOOLTIP_PREFIX + 'arr0w').addClass(TOOLTIP_PREFIX + 'bottom-arr0w')
                                                       .get()
                                                       );
}

function applyArrowPosition($el, tp, sel_ix, sel_iy, scale, oparams) {
    var tt_ix = tp[0];
    var cursor_type = tp[2];
    var cursor_width = 32;
    
    var absolute_selection_left_scroll = sel_ix + getBodyScrollLeft();
    var selection_absolute_width = oparams.width + oparams.pl + oparams.pr;
    var sel_center = absolute_selection_left_scroll + selection_absolute_width / 2;
    
    var cursorLeftMargin = sel_center - tt_ix - cursor_width * scale / 2;
    
    $el.find('.' + TOOLTIP_PREFIX + 'help-selected-wrap').addClass(TOOLTIP_PREFIX + 'has-' + cursor_type + '-arr0w');
    
    $el.find('.' + TOOLTIP_PREFIX + cursor_type + '-arr0w')
    .css('margin-left', cursorLeftMargin)
    .show();
}

// @returns null
function toggleLoadingInTooltip(ttid, is_loading) {
    //
    // remove previous loadings
    //
    $('.' + TOOLTIP_PREFIX + 'translate-loading').fadeOut(250, function () {
                                                          $(this).remove();
                                                          });
    
    if (is_loading) {
        var selection = _getSelection();
        var params = selection.toString()
        ? getSelectionParameters(selection)
        : last_call_args.selectionBB;
        
        var $dest = $('body');
        var is_netflix = false;
        
        if (document.location.href.indexOf('https://www.netflix.com/watch') > -1) {
            $dest = $('.nf-player-container');
            is_netflix = true;
        } else if ($dest.length === 0) {
            $dest = $('html');
        }
        
        var height = 70;
        var width = height;
        var x = params.left + params.width / 2 + window.pageXOffset - width / 2;
        var y;
        
        if (params.top + params.height + 10 + height > window.innerHeight || is_netflix) {
            y = params.top - 10 - height + window.pageYOffset; // loading is above
        } else {
            y = params.top + params.height + 10 + window.pageYOffset; // loading is below
        }
        
        $dest.append(
                     $('<div class="' + TOOLTIP_PREFIX + 'translate-loading"></div>')
                     .addClass(is_netflix || dark_mode ? TOOLTIP_PREFIX + 'dark-mode' : '')
                     .html('<div class="' + TOOLTIP_PREFIX + 'mate-loading"></div>')
                     .css({
                          top: y,
                          left: x
                          })
                     );
    }
}

function toggleOfflineInTooltip(ttid, is_offline) {
    if (is_offline) {
        $('.' + TOOLTIP_PREFIX + 'tooltip-' + ttid)
        .find('.' + TOOLTIP_PREFIX + 'offline')
        .css('display', 'flex');
    } else {
        $('.' + TOOLTIP_PREFIX + 'tooltip-' + ttid)
        .find('.' + TOOLTIP_PREFIX + 'offline')
        .fadeOut(250, function () {
                 $(this).remove();
                 });
    }
}

function setTooltipContents(id, code) {
    $('.' + TOOLTIP_PREFIX + 'content-layout-' + id).html(code);
    $('#' + TOOLTIP_PREFIX + 'tooltip-wrap *').addClass('TnITTtw-t');
}

const max_synonyms = 3;

var TRANSLITERATION_VARIANT = '\
<div class="<%=prefix%>variants-by-pos">\
<div class="<%=prefix%>variant-row <%=prefix%>translit-row">\
<div class="<%=prefix%>v-item <%=prefix%>translit-item <%=prefix%>no-border-bottom">\
<div class="<%=prefix%>v-pos <%=prefix%>translit-pos"><%=l_transliteration%></div>\
<div class="<%=prefix%>main-of-item <%=prefix%>translit-main"><%=transliteration%></div>\
</div>\
</div>\
</div>\
';

var IPA_VARIANT = '\
<div class="<%=prefix%>variants-by-pos">\
<div class="<%=prefix%>variant-row <%=prefix%>ipa-row">\
<div class="<%=prefix%>v-item <%=prefix%>ipa-item <%=prefix%>no-border-bottom">\
<div class="<%=prefix%>v-pos <%=prefix%>ipa-pos"><%=l_ipa%></div>\
<div class="<%=prefix%>main-of-item <%=prefix%>ipa-main"><%=ipa%></div>\
</div>\
</div>\
</div>\
';

var SINGLE_HTML = '\
<div class="<%=prefix%>original-wrap <%=prefix%>padded-single-translation">\
<div class="<%=prefix%>mv-text-part"><%=original%></div>\
<div class="<%=prefix%>ico-listen <%=prefix%>listen-butt0n <%=prefix%>listen-original" data-from="<%=from%>"></div>\
</div>\
<div class="<%=prefix%>padded-single-translation <%=prefix%>trans-wrap">\
<div class="<%=prefix%>tpart"><%=translation%></div>\
<div class="<%=prefix%>copy-translation-butt0n" data-to="<%=to%>"></div>\
<div class="<%=prefix%>ico-listen <%=prefix%>listen-butt0n <%=prefix%>listen-translation" data-to="<%=to%>"></div>\
</div>';

var positions = {
    '': 0,
noun: 1,
verb: 2,
adjective: 3,
adverb: 4,
pronoun: 5,
preposition: 6,
conjunction: 7,
interjection: 8,
abbreviation: 9,
phrase: 10,
suffix: 11,
auxiliaryverb: 12
};

var parts_of_speech = [];

for (var key in positions) {
    parts_of_speech.push(key);
}

function capitalize(s) {
    return !s ? '' : s[0].toUpperCase() + s.substr(1).toLowerCase();
}

var wrapper = {
TRANSLIT_TYPE: 1,
IPA_TYPE: 2,
SYNONYMS_TYPE: 3,
    
_singleWrap: function (translations, type, prefix, locales, lang) {
    prefix = prefix || '';
    
    if (type === wrapper.TRANSLIT_TYPE) {
        return show_translit && translations[4] ? compileString(TRANSLITERATION_VARIANT, {
                                                                prefix: prefix,
                                                                l_transliteration: locales.transliteration,
                                                                transliteration: translations[4],
                                                                to: lang
                                                                }) : '';
    } else if (type === wrapper.IPA_TYPE) {
        return show_ipa && translations[2] ? compileString(IPA_VARIANT, {
                                                           prefix: prefix,
                                                           l_ipa: locales.ipa,
                                                           ipa: translations[2],
                                                           to: lang
                                                           }) : '';
    } else if (type === wrapper.SYNONYMS_TYPE) {
        return compileString(SINGLE_HTML, {
                             prefix: prefix,
                             l_open: locales.open,
                             l_original: locales.original,
                             l_reversed: locales.reversed,
                             l_unpin: locales.unpin,
                             l_transliteration: locales.transliteration,
                             original: translations[1],
                             translation: translations[3],
                             transliteration: translations[4],
                             to: lang
                             });
    }
    
    return '';
},
    
_complexSingleWrap: function (translations, ov, prefix, locales, lang) {
    return this._singleWrap(translations, wrapper.SYNONYMS_TYPE, prefix, locales, lang)
    + this._singleWrap(translations, wrapper.TRANSLIT_TYPE, prefix, locales, lang)
    + this._singleWrap(translations, wrapper.IPA_TYPE, prefix, locales, lang);
},
    
_multiWrap: function (translations, onlyVariants, prefix, locales, lang) {
    var df_local, df_local_items;
    var df = document.createDocumentFragment();
    
    if (!prefix) {
        prefix = '';
    }
    
    if (show_translit && translations[4]) {
        $(df)
        .append($('<div>', {class: prefix + 'variant-row ' + prefix + 'translit-row'})
                .append($('<div>', {class: prefix + 'v-item ' + prefix + 'translit-item'})
                        .append($('<div>', {class: prefix + 'v-pos ' + prefix + 'translit-pos'})
                                .html(locales.transliteration))
                        .append($('<div>', {class: prefix + 'main-of-item ' + prefix + 'translit-main'})
                                .html(translations[4]))));
    }
    
    if (show_ipa && translations[2]) {
        $(df)
        .append($('<div>', {class: prefix + 'variant-row ' + prefix + 'ipa-row'})
                .append($('<div>', {class: prefix + 'v-item ' + prefix + 'ipa-item'})
                        .append($('<div>', {class: prefix + 'v-pos ' + prefix + 'ipa-pos'})
                                .html(locales.ipa))
                        .append($('<div>', {class: prefix + 'main-of-item ' + prefix + 'ipa-main'})
                                .html(translations[2]))));
    }
    
    for (var i = 0; i < translations[7].length; ++i) {
        if (!$.isArray(translations[7][i]) || translations[7][i].length === 0) {
            continue;
        }
        
        var len = translations[7][i].length;
        df_local = document.createDocumentFragment();
        
        $.each(translations[7][i], function (k, v) {
               df_local_items = document.createDocumentFragment();
               
               // Hotfix
               if (v[2] && typeof v[1] === 'string' && $.isArray(v[2])) {
               var t = v[1];
               v[1] = v[2];
               v[2] = t;
               }
               
               $.each(v[1] || [], function (k2, v2) {
                      if (k2 >= max_synonyms) return;
                      if (v2) {
                      $(df_local_items)
                      .append($('<div>', {class: prefix + 'synonym'}).html(v2))
                      .append(k2 < v[1].length - 1 && k2 < max_synonyms - 1 ? ', ' : '');
                      }
                      });
               
               var gender = '';
               
               if (v[2]) {
               gender = $('<div>', {class: prefix + 'gender'}).html(', ' + v[2]);
               }
               
               $(df_local)
               .append($('<div>', {class: prefix + 'v-item'})
                       .append($('<div>', {
                                 class: prefix + 'small-copy-button',
                                 'data-langto': lang
                                 }))
                       .append($('<div>', {
                                 class: prefix + 'listen-v-item',
                                 'data-langto': lang
                                 }))
                       .append($('<div>', {class: prefix + 'v-texts'})
                               .append($('<div>', {class: prefix + 'main-of-item'}).html(v[0]))
                               .append(gender))
                       .append($('<div>', {class: prefix + 'synonyms'}).append(df_local_items)));
               });
        
        var key = parts_of_speech[i];
        
        $(df).append(
                     $('<div>')
                     .addClass(prefix + 'variant-row')
                     .append(
                             $('<div>')
                             .addClass(prefix + 'v-pos')
                             .html(key ? getLocale('CommonUi_LangPart_' + capitalize(key)) : '')
                             .addClass(!key ? prefix + 'empty-pos' : '')
                             .get()
                             )
                     .append(
                             $('<div>')
                             .addClass(prefix + 'v-closest-wrap')
                             .append(df_local)
                             .get()
                             )
                     .get()
                     );
    }
    
    // Do not include different wrappers and a main variant to the final HTML code
    if (onlyVariants === true) {
        return $('<div>').append(df).html();
    }
    
    var bunch = $('<div>').addClass(prefix + 'variant-bunch-wrap').append(
                                                                          $('<div>')
                                                                          .addClass(prefix + 'vbw-inside-layout')
                                                                          .append(
                                                                                  $('<div>')
                                                                                  .addClass(prefix + 'original-wrap')
                                                                                  .append(
                                                                                          $('<div>')
                                                                                          .addClass(prefix + 'original')
                                                                                          .html('<div class="' + prefix + 'mv-text-part">' + translations[1] + '</div><div class="' + prefix + 'add-pb-butt0n"></div><div class="' + prefix + 'copy-butt0n"></div><div class="' + prefix + 'ico-listen ' + prefix + 'listen-butt0n ' + prefix + 'listen-original" data-from="<%=from%>">' +
                                                                                                '</div>')
                                                                                          .get()
                                                                                          )
                                                                                  .get()
                                                                                  )
                                                                          .append(
                                                                                  $('<div>')
                                                                                  .addClass(prefix + 'main-variant-wrap')
                                                                                  .append(
                                                                                          $('<div>')
                                                                                          .addClass(prefix + 'main-variant')
                                                                                          .html('<div class="' + prefix + 'mv-text-part">' + translations[3] + '</div><div class="' + prefix + 'copy-translation-butt0n"></div><div class="' + prefix + 'ico-listen ' + prefix + 'listen-butt0n ' + prefix + 'listen-translation" data-to="' + lang + '">' +
                                                                                                '</div>')
                                                                                          .get()
                                                                                          )
                                                                                  .get()
                                                                                  )
                                                                          .append(
                                                                                  $('<div>')
                                                                                  .addClass(prefix + 'variants-by-pos')
                                                                                  .append(df)
                                                                                  .get()
                                                                                  )
                                                                          .get()
                                                                          );
    
    return $('<div>').append(bunch.get()).html();
},
    
wrap: function (json) {
    return wrapper['_' + (json[0] ? 'multi' : 'complexSingle') + 'Wrap'](json, false, TOOLTIP_PREFIX, {
                                                                         original: getLocale('Kernel_Original'),
                                                                         translation: getLocale('Kernel_Translation'),
                                                                         ipa: getLocale('Kernel_IPA'),
                                                                         reversed: getLocale('Kernel_Reverse'),
                                                                         unpin: getLocale('Kernel_Unpin'),
                                                                         highlight: getLocale('Kernel_Highlight'),
                                                                         transliteration: getLocale('CommonUi_LangPart_Transliteration')
                                                                         }, json[6]);
}
};
